import os

from . import strings


def get_env_or_else(env, default=None):
    if isinstance(env, list):
        for val in env:
            value = os.getenv(val)
            if strings.is_not_blank(value):
                return value

    if isinstance(env, str):
        return os.getenv(env, default)

    return default



import os

import pytest
import yaml

from gcasc.utils.yaml_include import YamlIncluderConstructor

from .helpers import read_file, read_yaml

YamlIncluderConstructor.add_to_loader_class(loader_class=yaml.FullLoader,
                                            base_dir=os.path.dirname(os.path.realpath(__file__)) + '/data')


@pytest.fixture()
def file1():
    return read_yaml('yaml_include_f1.yml')


@pytest.fixture()
def file2():
    return read_yaml('yaml_include_f2.yml')


@pytest.fixture()
def file_txt():
    return read_file('yaml_include_txt.md')


def test_files_included_into_yaml(file1, file2, file_txt):
    # given
    file = 'yaml_include.yml'

    # when
    data = read_yaml(file)

    # then
    assert data['inc1'] == file1
    assert data['inc2'] == [file2, file_txt]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
from recommonmark.transform import AutoStructify

on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# -- Project information -----------------------------------------------------

project = 'GitLab Configuration as Code'
copyright = '2019, Hoffmann-La Roche'
author = 'Hoffmann-La Roche'

# The full version, including alpha/beta/rc tags
release = '0.1'

# -- General configuration ---------------------------------------------------

# The master toctree document.
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_markdown_tables",
    "recommonmark"
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "default"
if not on_rtd:  # only import and set the theme if we're building docs locally
    try:
        import sphinx_rtd_theme
        extensions.append("sphinx_rtd_theme")
        html_theme = "sphinx_rtd_theme"
    except ImportError:  # Theme not found, use default
        pass

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

highlight_language = "yaml"

source_suffix = ['.rst', '.md']

def setup(app):
    app.add_config_value('recommonmark_config', {
            # 'enable_auto_toc_tree': True,
            }, True)
    app.add_transform(AutoStructify)


# -*- coding: utf-8 -*-

import os.path

import yaml

try:
    from yaml import FullLoader
except ImportError:
    FullLoader = None

__all__ = ["YamlEnvConstructor"]


class YamlEnvConstructor:
    DEFAULT_TAG_NAME = "!env"

    def __call__(self, loader, node):
        args = []
        if isinstance(node, yaml.nodes.ScalarNode):
            args = [loader.construct_scalar(node)]
        elif isinstance(node, yaml.nodes.SequenceNode):
            args = loader.construct_sequence(node)
        else:
            raise TypeError("Un-supported YAML node {!r}".format(node))
        return YamlEnvConstructor.load(*args)

    @staticmethod
    def load(env):
        # type: (str)-> str
        splitted = env.split(":", 1)
        value = os.getenv(splitted[0], splitted[1] if splitted.__len__() == 2 else None)
        if value is None:
            raise RuntimeError(
                "Expected {0} environment variable, but value was not found in environment".format(
                    env
                )
            )
        return value

    @classmethod
    def add_to_loader_class(cls, loader_class=None, tag=None, **kwargs):
        # type: (type(yaml.Loader), str, **str)-> YamlEnvConstructor

        if tag is None:
            tag = ""
        tag = tag.strip()
        if not tag:
            tag = cls.DEFAULT_TAG_NAME
        if not tag.startswith("!"):
            raise ValueError('`tag` argument should start with character "!"')
        instance = cls(**kwargs)
        if loader_class is None:
            if FullLoader:
                yaml.add_constructor(tag, instance, FullLoader)
            else:
                yaml.add_constructor(tag, instance)
        else:
            yaml.add_constructor(tag, instance, loader_class)
        return instance



import os

import yaml
from gcasc.utils.yaml_include import YamlIncluderConstructor
from gcasc.utils.yaml_env import YamlEnvConstructor

YamlIncluderConstructor.add_to_loader_class(
    loader_class=yaml.FullLoader, base_dir=os.path.dirname(os.path.realpath(__file__))
)

YamlEnvConstructor.add_to_loader_class(loader_class=yaml.FullLoader)


class GitlabConfiguration(object):
    def __init__(self, config):
        # type: (dict)->GitlabConfiguration
        if config is None:
            raise RuntimeError("GitLab configuration is empty")

        if not isinstance(config, dict):
            raise RuntimeError("Configuration provided must be of dictionary type")

        self.config = config

    def get(self, configuration):
        return self.config.get(configuration)

    @staticmethod
    def from_file(path):
        # type: (str)->GitlabConfiguration
        with open(path) as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        return GitlabConfiguration(data)



import pytest
from pytest import mark
from unittest.mock import Mock, MagicMock
from gcasc import FeaturesConfigurer
from gcasc import Mode
from .helpers import read_yaml


@pytest.fixture()
def features_valid():
    return read_yaml('features_valid.yml')['features']


@pytest.fixture()
def features_invalid():
    return read_yaml('features_invalid.yml')['features']


@pytest.fixture()
def features_valid_canary():
    return read_yaml('features_valid_canary.yml')['features']


@pytest.fixture()
def features_valid_canaries():
    return read_yaml('features_valid_canaries.yml')['features']


def __mock_gitlab(features=[]):
    features_manager = Mock()
    features_manager.list.return_value = features
    return Mock(features=features_manager)


def test_features_configuration_valid(features_valid):
    # given
    configurer = FeaturesConfigurer(Mock(), features_valid, Mode.TEST_SKIP_VALIDATION)

    # when
    result = configurer.validate()

    # then
    assert not result.has_errors()


def test_features_configuration_invalid(features_invalid):
    # given
    configurer = FeaturesConfigurer(Mock(), features_invalid, Mode.TEST_SKIP_VALIDATION)

    # when
    result = configurer.validate()

    # then
    assert result.has_errors()
    assert result.errors.__len__() == 2
    assert result.has_error('name')
    assert result.has_error('value')


def test_existing_features_removed_before_applying():
    # given
    feature1 = Mock()
    feature2 = Mock()
    features = [feature1, feature2]
    gitlab = __mock_gitlab(features)
    configurer = FeaturesConfigurer(gitlab, [])

    # when
    configurer.configure()

    # then
    feature1.delete.assert_called_once()
    feature2.delete.assert_called_once()


def test_canaries_configured_when_in_config(features_valid_canary):
    # given
    feature = features_valid_canary[0]
    name = feature["name"]
    value = feature["value"]
    user = feature["users"][0]
    group = feature["groups"][0]
    project = feature["projects"][0]
    gitlab = __mock_gitlab()
    configurer = FeaturesConfigurer(gitlab, features_valid_canary)

    # when
    configurer.configure()

    # then
    gitlab.features.set.assert_any_call(name,
                                        value,
                                        feature_group=None,
                                        user=user)

    gitlab.features.set.assert_any_call(name,
                                        value,
                                        feature_group=None,
                                        group=group)

    gitlab.features.set.assert_any_call(name,
                                        value,
                                        feature_group=None,
                                        project=project)


def test_multiple_canaries_are_configured(features_valid_canaries):
    # given
    gitlab = __mock_gitlab()
    configurer = FeaturesConfigurer(gitlab, features_valid_canaries)

    # when
    configurer.configure()

    # then
    assert gitlab.features.set.call_count == 4


@mark.parametrize('mode', [Mode.TEST, Mode.TEST_SKIP_VALIDATION])
def test_configuration_not_applied_when_in_mode_other_than_not_apply(features_valid, mode):
    # given
    gitlab = __mock_gitlab()
    configurer = FeaturesConfigurer(gitlab, features_valid, mode)

    # when
    configurer.configure()

    # then
    assert gitlab.features.set.call_count == 0



from .base import Mode, UpdateOnlyConfigurer


class SettingsConfigurer(UpdateOnlyConfigurer):
    _NAME = "settings"

    def __init__(self, gitlab, settings, mode=Mode.APPLY):
        super().__init__("ApplicationSettings", gitlab, settings, mode=mode)

    def _load(self):
        return self.gitlab.settings.get()

    def validate(self):
        pass



from unittest.mock import Mock

import pytest

from gcasc import Mode, SettingsConfigurer

from .helpers import read_yaml


@pytest.fixture()
def settings_valid():
    return read_yaml("settings_valid.yml")["settings"]


def test_settings_not_updated_because_unchanged(settings_valid):
    # given
    gitlab = Mock()
    settings = Mock()
    settings.auto_devops_enabled = settings_valid["auto_devops_enabled"]
    settings.help_page_text = settings_valid["help_page"]["text"]
    settings.help_page_support_url = settings_valid["help_page"]["support_url"]
    settings.polling_interval_multiplier = settings_valid["polling_interval_multiplier"]
    settings.throttle_unauthenticated_enabled = settings_valid["throttle"]["unauthenticated"]["enabled"]
    settings.allowed_domains = settings_valid["allowed_domains"]
    gitlab.settings.get.return_value = settings
    configurer = SettingsConfigurer(gitlab, settings_valid)

    # when
    configurer.configure()

    # then
    gitlab.settings.get.assert_called_once()
    settings.save.assert_not_called()


def test_settings_modified(settings_valid):
    # given
    gitlab = Mock()
    settings = Mock()
    settings.auto_devops_enabled = settings_valid["auto_devops_enabled"]
    settings.help_page_text = "modified help page text"
    settings.help_page_support_url = settings_valid["help_page"]["support_url"]
    settings.polling_interval_multiplier = settings_valid["polling_interval_multiplier"]
    gitlab.settings.get.return_value = settings
    configurer = SettingsConfigurer(gitlab, settings_valid)

    # when
    saved = configurer.configure()

    # then
    gitlab.settings.get.assert_called_once()
    settings.save.assert_called_once()
    assert saved.help_page_text == settings_valid["help_page"]["text"]


def test_invalid_configuration_options_are_skipped(settings_valid):
    # given
    gitlab = Mock()
    settings = Mock()
    settings.help_page_text = settings_valid["help_page"]["text"]
    gitlab.settings.get.return_value = settings
    configurer = SettingsConfigurer(gitlab, settings_valid)

    # when
    configurer.configure()

    # then
    gitlab.settings.get.assert_called_once()


@pytest.mark.parametrize("mode", [Mode.TEST, Mode.TEST_SKIP_VALIDATION])
def test_no_changes_in_not_apply_mode(mode, settings_valid):
    # given
    text = "modified help page text"
    gitlab = Mock()
    settings = Mock()
    settings.help_page_text = text
    gitlab.settings.get.return_value = settings
    configurer = SettingsConfigurer(gitlab, settings_valid, mode)

    # when
    saved = configurer.configure()

    # then
    gitlab.settings.get.assert_called_once()
    settings.save.assert_not_called()
    assert saved.help_page_text == text



from .utils import logger

from .base import Configurer, Mode, ValidationResult

logger = logger.get_logger("Features")


class FeaturesConfigurer(Configurer):
    _NAME = "features"

    def __init__(
        self, gitlab, features, mode=Mode.APPLY
    ):  # type: (gitlab.Gitlab, dict, Mode)->FeaturesConfigurer
        super().__init__(gitlab, features, mode=mode)

    def configure(self):
        logger.info("Configuring GitLab Features")
        self.__remove_existing()

        for feature in self.config:
            name = feature["name"]
            value = feature["value"]
            if self.mode == Mode.APPLY:
                self.__apply(name, value, feature)
            logger.info("Configured: %s => %s", name, value)
        return self.gitlab.features.list()

    def __set_canary(
        self, canary_name, canaries, feature_name, feature_value, feature_group
    ):
        if canaries:
            for canary_value in canaries:
                logger.info("Configuring canary: %s => %s", canary_name, canary_value)
                self.gitlab.features.set(
                    feature_name,
                    feature_value,
                    feature_group=feature_group,
                    **{canary_name: canary_value}
                )
            return True
        return False

    def __apply(self, name, value, feature):
        feature_group = feature.get("feature_group")
        users = feature.get("users")
        is_set = self.__set_canary("user", users, name, value, feature_group)
        groups = feature.get("groups")
        is_set = (
            self.__set_canary("group", groups, name, value, feature_group) or is_set
        )
        projects = feature.get("projects")
        is_set = (
            self.__set_canary("project", projects, name, value, feature_group) or is_set
        )
        if not is_set:  # set feature globally
            self.gitlab.features.set(name, value, feature_group=feature_group)

    def __remove_existing(self):
        if self.mode == Mode.APPLY:
            features = self.gitlab.features.list()
            [feature.delete() for feature in features]

    def validate(self):  # type: () -> ValidationResult
        errors = ValidationResult()
        for idx, feature in enumerate(self.config):
            name = feature.get("name")
            if not name:
                errors.add("feature[{0}] must have name property", str(idx))
            else:
                name = "{0}.{1}".format(idx, name)
            if feature.get("value") is None:
                errors.add("feature[{0}] must have value property", name)
        return errors



import logging
import os

level_str = os.getenv("LOGGING_LEVEL", "INFO")
level = logging.getLevelName(level_str)

logging.basicConfig(format="[%(levelname)s] [%(name)s] %(message)s", level=level)


def get_logger(name="gcasc"):
    return Logger(name)


def _apply_masking(value, mask=False):
    return "***" if mask else value


class Logger(object):
    def __init__(self, name):
        self.logger = logging.getLogger(name)

    def log_update(self, field, old_value, new_value, masked=False):
        self.debug(
            "Changed value of %s from %s to %s",
            field,
            _apply_masking(old_value, masked),
            _apply_masking(new_value, masked),
        )
        pass

    def log_create(self, value, field=None, masked=False):
        value_to_log = _apply_masking(value, masked)
        if field is None:
            self.debug("Created new object: %s", value_to_log)
        else:
            self.debug("Created new field %s with value %s", field, value_to_log)
        pass

    def log_delete(self, field):
        self.debug("Deleted field or object: %s", field)
        pass

    def debug(self, message, *args):
        self.logger.debug(message, *args)

    def error(self, message, *args):
        self.logger.error(message, *args)

    def warn(self, message, *args):
        self.logger.warning(message, *args)

    def info(self, message, *args):
        self.logger.info(message, *args)



# -*- coding: utf-8 -*-

# It is based on https://github.com/tanbro/pyyaml-include, but contains added handling of non yaml file types as strings

"""
Include YAML files within YAML
"""

import io
import os.path
import re
from glob import iglob
from sys import version_info

import yaml

try:
    from yaml import FullLoader
except ImportError:
    FullLoader = None

__all__ = ["YamlIncluderConstructor"]

PYTHON_MAYOR_MINOR = "{0[0]}.{0[1]}".format(version_info)

WILDCARDS_REGEX = re.compile(r"^.*(\*|\?|\[!?.+\]).*$")


class YamlIncluderConstructor:
    """The `include constructor` for PyYAML Loaders

    Call :meth:`add_to_loader_class` or :meth:`yaml.Loader.add_constructor` to add it into loader.

    In YAML files, use ``!include`` to load other YAML files as below::

        !include [dir/**/*.yml, true]

    or::

        !include {pathname: dir/abc.yml, encoding: utf-8}

    """

    DEFAULT_ENCODING = "utf-8"
    DEFAULT_TAG_NAME = "!include"

    def __init__(self, base_dir=None, encoding=None):
        # type:(str, str)->YamlIncluderConstructor
        """
        :param str base_dir: Base directory where search including YAML files

            :default: ``None``:  include YAML files from current working directory.

        :param str encoding: Encoding of the YAML files

            :default: ``None``:  Not specified
        """
        self._base_dir = base_dir
        self._encoding = encoding

    def __call__(self, loader, node):
        args = []
        kwargs = {}
        if isinstance(node, yaml.nodes.ScalarNode):
            args = [loader.construct_scalar(node)]
        elif isinstance(node, yaml.nodes.SequenceNode):
            args = loader.construct_sequence(node)
        elif isinstance(node, yaml.nodes.MappingNode):
            kwargs = loader.construct_mapping(node)
        else:
            raise TypeError("Un-supported YAML node {!r}".format(node))
        return self.load(loader, *args, **kwargs)

    @property
    def base_dir(self):  # type: ()->str
        """Base directory where search including YAML files

        :rtype: str
        """
        return self._base_dir

    @base_dir.setter
    def base_dir(self, value):  # type: (str)->None
        self._base_dir = value

    @property
    def encoding(self):  # type: ()->str
        """Encoding of the YAML files

        :rtype: str
        """
        return self._encoding

    @encoding.setter
    def encoding(self, value):  # type: (str)->None
        self._encoding = value

    @staticmethod
    def is_yaml(path):
        _, extension = os.path.splitext(path)
        return True if extension in [".yml", "yaml"] else False

    def load(self, loader, pathname, recursive=False, encoding=None):
        """Once add the constructor to PyYAML loader class,
        Loader will use this function to include other YAML fils
        on parsing ``"!include"`` tag

        :param loader: Instance of PyYAML's loader class
        :param str pathname: pathname can be either absolute (like /usr/src/Python-1.5/Makefile) or relative (like ../../Tools/*/*.gif), and can contain shell-style wildcards

        :param bool recursive: If recursive is true, the pattern ``"**"`` will match any files and zero or more directories and subdirectories. If the pattern is followed by an os.sep, only directories and subdirectories match.

            Note:
             Using the ``"**"`` pattern in large directory trees may consume an inordinate amount of time.

        :param str encoding: YAML file encoding

            :default: ``None``: Attribute :attr:`encoding` or constant :attr:`DEFAULT_ENCODING` will be used to open it

        :return: included YAML file, in Python data type

        .. warning:: It's called by :mod:`yaml`. Do NOT call it yourself.
        """
        if not encoding:
            encoding = self._encoding or self.DEFAULT_ENCODING
        if self._base_dir:
            pathname = os.path.join(self._base_dir, pathname)
        if re.match(WILDCARDS_REGEX, pathname):
            result = []
            if PYTHON_MAYOR_MINOR >= "3.5":
                iterator = iglob(pathname, recursive=recursive)
            else:
                iterator = iglob(pathname)
            for path in iterator:
                if os.path.isfile(path):
                    with io.open(
                        path, encoding=encoding
                    ) as fp:  # pylint:disable=invalid-name
                        result.append(yaml.load(fp, type(loader)))
            return result
        with open(pathname, encoding=encoding) as fp:  # pylint:disable=invalid-name
            if self.is_yaml(pathname):
                return yaml.load(fp, type(loader))
            return fp.read()

    @classmethod
    def add_to_loader_class(cls, loader_class=None, tag=None, **kwargs):
        # type: (type(yaml.Loader), str, **str)-> YamlIncluderConstructor
        """
        Create an instance of the constructor, and add it to the YAML `Loader` class

        :param loader_class: The `Loader` class add constructor to.

            .. attention:: This parameter **SHOULD** be a **class type**, **NOT** object.

            It's one of following:

                - :class:`yaml.BaseLoader`
                - :class:`yaml.UnSafeLoader`
                - :class:`yaml.SafeLoader`
                - :class:`yaml.Loader`
                - :class:`yaml.FullLoader`
                - :class:`yaml.CBaseLoader`
                - :class:`yaml.CUnSafeLoader`
                - :class:`yaml.CSafeLoader`
                - :class:`yaml.CLoader`
                - :class:`yaml.CFullLoader`

            :default: ``None``:

                - When :mod:`pyyaml` 3.*: Add to PyYAML's default `Loader`
                - When :mod:`pyyaml` 5.*: Add to `FullLoader`

        :type loader_class: type

        :param str tag: Tag's name of the include constructor.

          :default: ``""``: Use :attr:`DEFAULT_TAG_NAME` as tag name.

        :param kwargs: Arguments passed to construct function

        :return: New created object
        :rtype: YamlIncluderConstructor
        """
        if tag is None:
            tag = ""
        tag = tag.strip()
        if not tag:
            tag = cls.DEFAULT_TAG_NAME
        if not tag.startswith("!"):
            raise ValueError('`tag` argument should start with character "!"')
        instance = cls(**kwargs)
        if loader_class is None:
            if FullLoader:
                yaml.add_constructor(tag, instance, FullLoader)
            else:
                yaml.add_constructor(tag, instance)
        else:
            yaml.add_constructor(tag, instance, loader_class)
        return instance



import os

import pytest
import yaml

from gcasc.utils.yaml_env import YamlEnvConstructor

from .helpers import read_yaml

YamlEnvConstructor.add_to_loader_class(loader_class=yaml.FullLoader)

ENV1_VAL = 'envvar1'
ENV2_VAL = 'envvar2\nmultiline\ttabbed'


@pytest.fixture(autouse=True)
def prepare_environment():
    # before
    os.environ['ENV1'] = ENV1_VAL
    os.environ['ENV2'] = ENV2_VAL
    # test
    yield
    # after
    os.unsetenv('ENV1')
    os.unsetenv('ENV2')


def test_environment_injected_into_yaml():
    # given
    file = 'yaml_env.yml'

    # when
    data = read_yaml(file)

    # then
    assert data['env1'] == ENV1_VAL
    assert data['env_list'] == [ENV1_VAL, 'someval', ENV2_VAL]



import pytest

from gcasc.utils import validators


@pytest.mark.parametrize('date', ['2001-01-01', '2050-12-31'])
def test_date_success_validation(date):
    # expect
    assert validators.validate_date(date)


@pytest.mark.parametrize('date', ['2001-jan-01', '31-12-2050', '2050 10 12', ''])
def test_date_failure_validation(date):
    # expect
    assert not validators.validate_date(date)



import os
from pathlib import Path

import gitlab
import requests

import gcasc.utils.os as uos
from .utils import logger as logging

from .base import Mode
from .appearance import AppearanceConfigurer
from .license import LicenseConfigurer
from .features import FeaturesConfigurer
from .settings import SettingsConfigurer
from .config import GitlabConfiguration

GITLAB_CLIENT_CONFIG_FILE = ["GITLAB_CLIENT_CONFIG", "GITLAB_CLIENT_CONFIG_FILE"]
GITLAB_CLIENT_CERTIFICATE = ["GITLAB_CLIENT_CERT", "GITLAB_CLIENT_CERTIFICATE"]
GITLAB_CLIENT_KEY = "GITLAB_CLIENT_KEY"
GITLAB_CLIENT_URL = "GITLAB_CLIENT_URL"
GITLAB_CLIENT_API_VERSION = "GITLAB_CLIENT_API_VERSION"
GITLAB_CLIENT_TOKEN = "GITLAB_CLIENT_TOKEN"
GITLAB_CLIENT_SSL_VERIFY = "GITLAB_CLIENT_SSL_VERIFY"
GITLAB_CONFIG_FILE = ["GITLAB_CONFIG_FILE", "GITLAB_CONFIG_PATH"]

GITLAB_CONFIG_FILE_DEFAULT_PATHS = [
    "/etc/python-gitlab.cfg",
    "/etc/gitlab.cfg",
    "~/.python-gitlab.cfg",
    "~/.gitlab.cfg",
]

GITLAB_MODE = "GITLAB_MODE"

logger = logging.get_logger()

configurers = [
    SettingsConfigurer,
    LicenseConfigurer,
    AppearanceConfigurer,
    FeaturesConfigurer,
]


def create_all_configurers(gitlab, config, mode):
    return map(
        lambda configurer: configurer(gitlab, config.get(configurer._NAME), mode),
        configurers,
    )


class GitlabConfigurationAsCode(object):
    def __init__(self, configurers=None):
        # type: ()->GitlabConfigurationAsCode
        self.mode = Mode[uos.get_env_or_else(GITLAB_MODE, Mode.APPLY.name)]
        if self.mode == Mode.TEST:
            logger.info("TEST MODE ENABLED. NO CHANGES WILL BE APPLIED")
        path = uos.get_env_or_else(
            GITLAB_CONFIG_FILE, "{0}/gitlab.yml".format(os.getcwd())
        )
        self.gitlab = init_gitlab_client()
        self.config = GitlabConfiguration.from_file(path)
        self.configurers = (
            configurers
            if configurers
            else create_all_configurers(self.gitlab, self.config, self.mode)
        )
        version, revision = self.gitlab.version()
        logger.info("GitLab version: %s, revision: %s", version, revision)

    def configure(self, target=None):
        if target is None:
            for configurer in self.configurers:
                if configurer.config is None:
                    logger.info(
                        "Skipping configurer %s because it does not have any configuration to apply",
                        configurer.__class__.__name__,
                    )
                    continue
                configurer.configure()


def init_gitlab():
    # type: ()->gitlab.Gitlab
    config_path = __find_gitlab_connection_config_file()
    config = (
        gitlab.config.GitlabConfigParser(gitlab_id="global", config_files=[config_path])
        if config_path is not None
        else None
    )

    token = getattr(config, "private_token", None) or uos.get_env_or_else(
        GITLAB_CLIENT_TOKEN
    )
    if token is None:
        raise ClientInitializationError(
            "GitLab token was not provided. It must be defined in {0} environment variable or config file".format(
                GITLAB_CLIENT_TOKEN
            )
        )

    url = getattr(config, "url", None) or uos.get_env_or_else(
        GITLAB_CLIENT_URL, "https://gitlab.com"
    )
    # ssl_verify is always set by GitlabConfigParser, using inline if to handle `false` value read from config file
    ssl_verify = (
        config.ssl_verify
        if config
        else uos.get_env_or_else(GITLAB_CLIENT_SSL_VERIFY, True)
    )
    api_version = getattr(config, "api_version", None) or uos.get_env_or_else(
        GITLAB_CLIENT_API_VERSION, "4"
    )

    return gitlab.Gitlab(
        url=url, private_token=token, ssl_verify=ssl_verify, api_version=api_version
    )


def init_gitlab_client():
    # type: ()->gitlab.Gitlab
    logger.info("Initializing GitLab client")
    logger.info("Trying to initialize GitLab client...")
    client = init_gitlab()

    if client is None:
        raise ClientInitializationError(
            "Unable to initialize GitLab client due to missing configuration either in "
            "config file or environment vars"
        )

    __init_session(client)
    return client


def __find_gitlab_connection_config_file():
    # type: ()->str
    config_path = uos.get_env_or_else(GITLAB_CLIENT_CONFIG_FILE)
    if config_path is not None:
        if not __check_file_exists(config_path, "GitLab Client"):
            logger.error(
                "Configuration file was not found under path %s, which was defined in %s env variable."
                "Provide path to existing file or remove this variable and configure client"
                "using environment variables instead of configuration file",
                config_path,
                GITLAB_CLIENT_CONFIG_FILE,
            )
            return None
    else:
        for path in GITLAB_CONFIG_FILE_DEFAULT_PATHS:
            if __check_file_exists(path, "GitLab Client"):
                config_path = path
                break
    return config_path


def __check_file_exists(path, file_context=""):
    # type: (str, str)->bool
    config = Path(path)
    if not config.exists():
        logger.error(
            "[%s] File under %a does not exist. Provide valid path.", file_context, path
        )
        return False

    if config.is_dir():
        logger.error(
            "[%s] Directory was found under %s instead of file. Provide valid path to file.",
            file_context,
            path,
        )
        return False
    return True


def __init_session(gitlab):
    # type: (gitlab.Gitlab)->()
    certificate = uos.get_env_or_else(GITLAB_CLIENT_CERTIFICATE)
    key = uos.get_env_or_else(GITLAB_CLIENT_KEY)
    if certificate is None and key is None:
        return
    elif certificate is None and key is not None:
        pass
    elif certificate is not None and key is None:
        pass
    else:
        check_config = __check_file_exists(
            certificate, "Client Certificate"
        ) and __check_file_exists(key, "Client Key")
        if not check_config:
            raise ClientInitializationError(
                "GitLab client authentication env vars were provided, but point to incorrect file(s)"
            )
        session = requests.Session()
        session.cert = (certificate, key)
        gitlab.session = session


class ClientInitializationError(Exception):
    pass



import os

import pytest
from mock import Mock, patch

from gcasc import ClientInitializationError, GitlabConfigurationAsCode, Mode
from tests import helpers

GITLAB_CLIENT_CONFIG_FILE = ["GITLAB_CLIENT_CONFIG", "GITLAB_CLIENT_CONFIG_FILE"]
GITLAB_CLIENT_CERTIFICATE = ["GITLAB_CLIENT_CERT", "GITLAB_CLIENT_CERTIFICATE"]
GITLAB_CLIENT_KEY = "GITLAB_CLIENT_KEY"
GITLAB_CLIENT_URL = "GITLAB_CLIENT_URL"
GITLAB_CLIENT_API_VERSION = "GITLAB_CLIENT_API_VERSION"
GITLAB_CLIENT_TOKEN = "GITLAB_CLIENT_TOKEN"
GITLAB_CLIENT_SSL_VERIFY = "GITLAB_CLIENT_SSL_VERIFY"
GITLAB_CONFIG_FILE = ["GITLAB_CONFIG_FILE", "GITLAB_CONFIG_PATH"]

GITLAB_MODE = "GITLAB_MODE"

GITLAB_CONFIG_FILE_DEFAULT_PATHS = [
    "/etc/python-gitlab.cfg",
    "/etc/gitlab.cfg",
    "~/.python-gitlab.cfg",
    "~/.gitlab.cfg",
]


def __mock_gitlab(gitlab_class_mock):
    gitlab = Mock()
    gitlab.version.return_value = ('test', 'test')
    gitlab_class_mock.from_config.return_value = gitlab
    gitlab_class_mock.return_value = gitlab
    return gitlab


@pytest.fixture(scope="class", autouse=True)
def configure_shared_environment(request):
    os.environ['GITLAB_CONFIG_FILE'] = helpers.get_file_path('gitlab.yml')

    gitlab_patch = patch('gitlab.Gitlab')
    gitlab_mock_class = gitlab_patch.__enter__()
    __mock_gitlab(gitlab_mock_class)

    def unpatch():
        gitlab_patch.__exit__()

    def clean_env():
        for key, value in os.environ.items():
            if key.startswith('GITLAB_'):
                del os.environ[key]

    request.addfinalizer(unpatch)
    request.addfinalizer(clean_env)


@patch('gitlab.Gitlab')
def test_gitlab_client_created_from_config_file(gitlab_class_mock):
    # given
    config_path = helpers.get_file_path('gitlab_config_valid.cfg')
    os.environ['GITLAB_CLIENT_CONFIG_FILE'] = config_path
    __mock_gitlab(gitlab_class_mock)

    # when
    GitlabConfigurationAsCode()

    # then
    gitlab_class_mock.assert_called_once_with(private_token='my_token', url='https://my.gitlab.com',
                                              ssl_verify=False, api_version='4')


@patch('gitlab.Gitlab')
def test_gitlab_client_created_from_environment(gitlab_class_mock):
    os.environ['GITLAB_CLIENT_TOKEN'] = 'token'
    os.environ['GITLAB_CLIENT_URL'] = 'url'
    os.environ['GITLAB_CLIENT_API_VERSION'] = 'api_version'
    os.environ['GITLAB_CLIENT_SSL_VERIFY'] = 'ssl_verify'
    __mock_gitlab(gitlab_class_mock)

    # when
    GitlabConfigurationAsCode()

    # then
    gitlab_class_mock.assert_called_once_with(private_token='token', url='url', ssl_verify='ssl_verify',
                                              api_version='api_version')

@patch('gitlab.Gitlab')
def test_gitlab_client_created_from_file_and_environment(gitlab_class_mock):
    # given
    config_path = helpers.get_file_path('gitlab_config_invalid.cfg')
    os.environ['GITLAB_CLIENT_CONFIG_FILE'] = config_path
    os.environ['GITLAB_CLIENT_TOKEN'] = 'token'
    __mock_gitlab(gitlab_class_mock)

    # when
    GitlabConfigurationAsCode()

    # then
    gitlab_class_mock.assert_called_once_with(private_token='token', url='https://my.gitlab.com',
                                              ssl_verify=True, api_version='4')


def test_gitlab_config_loaded_from_file():
    # given
    os.environ['GITLAB_CLIENT_TOKEN'] = 'some_token'

    # when
    gcasc = GitlabConfigurationAsCode()

    # then
    assert gcasc.config.config == helpers.read_yaml('gitlab.yml')


def test_session_initialized_when_config_provided():
    certificate = helpers.get_file_path('dummycert.crt')
    key = helpers.get_file_path('dummykey.key')
    os.environ['GITLAB_CLIENT_TOKEN'] = 'token'
    os.environ['GITLAB_CLIENT_CERTIFICATE'] = certificate
    os.environ['GITLAB_CLIENT_KEY'] = key

    gcasc = GitlabConfigurationAsCode()

    assert gcasc.gitlab.session.cert == (certificate, key)


def test_error_raised_when_unable_to_create_gitlab_client():
    with pytest.raises(ClientInitializationError) as error:
        GitlabConfigurationAsCode()

        assert "config file" in error
        assert "environment variables" in error


def test_apply_mode_is_used_when_not_provided():
    os.environ['GITLAB_CLIENT_TOKEN'] = 'token'

    gcasc = GitlabConfigurationAsCode()

    assert gcasc.mode == Mode.APPLY



from abc import ABC
from enum import Enum

from gitlab import Gitlab
from .utils import logger


class Mode(Enum):
    TEST_SKIP_VALIDATION = -2
    TEST = -1
    APPLY = 0


class Configurer(ABC):
    _NAME = None

    def __init__(
        self, gitlab, config, mode=Mode.APPLY
    ):  # type: (Gitlab, dict, Mode)->any

        if not self._NAME:
            raise RuntimeError(
                "Class property _NAME must be defined! It will be expected "
                "to exist in configuration file as a key"
            )

        self.gitlab = gitlab
        self.config = config
        self.mode = mode
        if self.mode != Mode.TEST_SKIP_VALIDATION:
            self._validate()

    def configure(self):
        pass

    def validate(self):  # type: () -> ValidationResult
        pass

    def _get(self, property):
        return self.config.get(property)

    def _validate(self):
        if self.gitlab is None:
            raise RuntimeError("GitLab client is not provided")
        if self.config is not None:
            result = self.validate()
            if result and result.has_errors():
                error = ValidationError.from_validation_result(result)
                raise error


class UpdateOnlyConfigurer(Configurer):
    def __init__(self, name, gitlab, config, mode=Mode.APPLY):
        self.name = name
        self.logger = logger.get_logger(name)
        super().__init__(gitlab, config, mode=mode)

    def _save(self, data):
        data.save()

    def _load(self):
        raise NotImplementedError("")

    def configure(self):
        self.logger.info("Configuring %s", self.name)
        data = self._load()
        changes = self._update_setting(data, self.config)
        self.logger.info("Found %s changed values", changes)
        if changes != 0:

            self.logger.info("Applying changes...")
            if self.mode == Mode.APPLY:
                self._save(data)
            else:
                self.logger.info("No changes will be applied due to test mode enabled")
        else:
            self.logger.info("Nothing to do")
        return data

    def _update_setting(
        self, current, new, changes=0, prefix=""
    ):  # type: (dict, dict, int, str)->int
        for key, value in new.items():
            if isinstance(value, dict):
                changes += self._update_setting(
                    current, value, changes, "{0}{1}_".format(prefix, key)
                )
                continue

            prefixed_key = "{0}{1}".format(prefix, key)

            self.logger.debug("Checking %s", prefixed_key)
            if hasattr(current, prefixed_key):
                current_value = getattr(current, prefixed_key)
                if current_value != value:
                    changes += 1
                    if current_value is None:
                        self.logger.info("Set: %s = %s", prefixed_key, value)
                    elif value is None:
                        self.logger.info("Unset: %s = %s", prefixed_key, current_value)
                    else:
                        self.logger.info(
                            "Updated %s: %s => %s", prefixed_key, current_value, value
                        )
                    if self.mode == Mode.APPLY:
                        setattr(current, prefixed_key, value)
            else:
                self.logger.warn(
                    "Invalid configuration option: %s. Skipping...", prefixed_key
                )
        return changes


class ValidationResult(object):
    def __init__(self, errors=None):
        self.errors = [] if errors is None else errors

    def add(self, error, *args):  # type: (str, *str) -> ()
        message = error.format(*args) if args else error
        self.errors.append(message)

    def get(self):
        return self.errors

    def get_as_string(self):
        return "\n".join(self.errors)

    def has_errors(self):
        return self.errors.__len__() > 0

    def has_error(self, error):
        for e in self.errors:
            if error in e:
                return True
        return False


class ValidationError(RuntimeError):
    @staticmethod
    def from_validation_result(validation_result):
        if validation_result is None:
            return ValidationError("Validation failed, but no errors provided")
        errors = validation_result.get_as_string()
        return ValidationError(
            "Configuration validation failed with errors:\n{0}".format(errors)
        )



from .base import UpdateOnlyConfigurer, Mode


class AppearanceConfigurer(UpdateOnlyConfigurer):
    _NAME = "appearance"

    def __init__(self, gitlab, settings, mode=Mode.APPLY):
        super().__init__("Appearance", gitlab, settings, mode=mode)

    def _load(self):
        return self.gitlab.appearance.get()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup


def clean_requirements(requirements):
    return [req for req in requirements if not req.startswith('-r')]


with open("README.md", "r") as readme_file:
    readme = readme_file.read()

with open("requirements.txt", "r") as reqs_file:
    requirements = reqs_file.readlines()
    requirements = clean_requirements(requirements)

with open("test-requirements.txt", "r") as test_reqs_file:
    test_requirements = test_reqs_file.readlines()
    test_requirements = clean_requirements(test_requirements)

setup(
    name="gitlab-configuration-as-code",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description="Manage GitLab configuration as code",
    long_description_content_type='text/markdown',
    long_description=readme,
    author="Mateusz Filipowicz",
    author_email="mateusz.filipowicz@roche.com",
    license="Apache-2.0",
    url="https://github.com/Roche/gitlab-configuration-as-code",
    keywords=['gitlab', 'configuration-as-code'],
    packages=find_packages(),
    install_requires=requirements,
    tests_require=test_requirements,
    entry_points={"console_scripts": ["gcasc = gcasc.bin.gcasc:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import datetime

from .utils import logger
import gcasc.utils.validators as validators

from .base import Configurer, Mode, ValidationResult

logger = logger.get_logger("License")


class LicenseConfigurer(Configurer):
    _NAME = "license"

    def __init__(
        self, gitlab, license, mode=Mode.APPLY
    ):  # type: (gitlab.Gitlab, dict, Mode)->LicenseConfigurer
        super().__init__(gitlab, license, mode=mode)

    def configure(self):
        logger.info("Configuring GitLab licenses")
        license = self.gitlab.get_license()
        if not self.__check_if_same_license(license) and self.mode == Mode.APPLY:
            license = self._update_license()
        logger.info(
            "Current license:\nplan: %s\nstarts_at: %s\nexpires_at: %s\nuser_limit: %s",
            self.__get_plan(license),
            self.__get_starts_at(license),
            self.__get_expires_at(license),
            self.__get_user_limit(license),
        )
        return license

    def __get(self, field, license=None):  # type: (str, dict) -> str
        license_config = license if license is not None else self.config
        return license_config.get(field)

    def __get_date(self, field, license=None):
        date = self.__get(field, license)
        return date.strftime("%Y-%m-%d") if isinstance(date, datetime.date) else date

    def __get_starts_at(self, license=None):  # type: (dict) -> str
        return self.__get_date("starts_at", license)

    def __get_expires_at(self, license=None):  # type: (dict) -> str
        return self.__get_date("expires_at", license)

    def __get_plan(self, license=None):  # type: (dict) -> str
        return self.__get("plan", license)

    def __get_user_limit(self, license=None):  # type: (dict) -> str
        return self.__get("user_limit", license)

    def __get_data(self, license=None):  # type: (dict) -> str
        return self.__get("data", license)

    def __check_if_same_license(self, license):
        return (
            self.__get_starts_at() == self.__get_starts_at(license)
            and self.__get_expires_at() == self.__get_expires_at(license)
            and self.__get_plan() == self.__get_plan(license)
            and self.__get_user_limit() == self.__get_user_limit(license)
        )

    def _update_license(self):
        logger.info("Updating GitLab license...")
        return self.gitlab.set_license(self.__get_data())

    def validate(self):  # type: () -> ValidationResult
        errors = ValidationResult()
        if not self.__get_starts_at():
            errors.add("License must have starts_at property set in format yyyy-MM-dd")
        elif not validators.validate_date(self.__get_starts_at(), "%Y-%m-%d"):
            errors.add(
                "starts_at license property must follow format yyyy-MM-dd, e.g. 2019-03-28"
            )
        if not self.__get_expires_at():
            errors.add("License must have expires_at property set in format yyyy-MM-dd")
        elif not validators.validate_date(self.__get_expires_at(), "%Y-%m-%d"):
            errors.add(
                "expires_at license property must follow format yyyy-MM-dd, e.g. 2019-03-28"
            )
        if not self.__get_plan():
            errors.add("License must have plan property set")
        elif not self.__get_plan() in ["starter", "premium", "ultimate"]:
            errors.add(
                "License plan property was '{0}', but should be one of: starter, premium or ultimate",
                self.__get_plan(),
            )
        if not self.__get_user_limit():
            errors.add("License must have user_limit property set in format")
        if not self.__get_data():
            errors.add(
                "License must have data property configured containing license itself"
            )
        return errors



from unittest.mock import Mock

import pytest

from gcasc import AppearanceConfigurer

from .helpers import read_yaml


@pytest.fixture()
def appearance_valid():
    return read_yaml('appearance_valid.yml')['appearance']


def test_appearance_not_updated_because_unchanged(appearance_valid):
    # given
    gitlab = Mock()
    appearance = Mock()
    appearance.title = appearance_valid['title']
    appearance.description = appearance_valid['description']
    appearance.header_message = appearance_valid['header']['message']
    gitlab.appearance.get.return_value = appearance
    configurer = AppearanceConfigurer(gitlab, appearance_valid)

    # when
    configurer.configure()

    # then
    gitlab.appearance.get.assert_called_once()
    appearance.save.assert_not_called()


def test_appearance_modified(appearance_valid):
    # given
    gitlab = Mock()
    appearance = Mock()
    appearance.title = appearance_valid['title']
    appearance.description = 'modified description'
    appearance.header_message = appearance_valid['header']['message']
    gitlab.appearance.get.return_value = appearance
    configurer = AppearanceConfigurer(gitlab, appearance_valid)

    # when
    saved = configurer.configure()

    # then
    gitlab.appearance.get.assert_called_once()
    appearance.save.assert_called_once()
    assert saved.description == appearance_valid['description']



import os

import yaml

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

def __context():
    dir = os.getcwd()
    if dir.endswith('tests'):
        return dir
    return '{0}/tests'.format(dir)


CONTEXT = __context()


def get_file_path(file):
    return '{0}/data/{1}'.format(CONTEXT, file)


def read_file(file):
    with open(get_file_path(file)) as f:
        data = f.read()
    return data


def read_yaml(file):
    with open(get_file_path(file)) as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
    return data


def read_yaml_from_string(str):
    return yaml.load(StringIO(str), Loader=yaml.FullLoader)



import pytest

from unittest.mock import Mock
from gcasc import LicenseConfigurer
from gcasc import Mode
from .helpers import read_yaml


@pytest.fixture()
def license_valid():
    return read_yaml('license_valid.yml')['license']


@pytest.fixture()
def license_invalid():
    return read_yaml('license_invalid_1.yml')['license']


def test_license_configuration_valid(license_valid):
    # given
    configurer = LicenseConfigurer(Mock(), license_valid, Mode.TEST_SKIP_VALIDATION)

    # when
    result = configurer.validate()

    # then
    assert not result.has_errors()


def test_license_configuration_invalid(license_invalid):
    # given
    configurer = LicenseConfigurer(Mock(), license_invalid, Mode.TEST_SKIP_VALIDATION)

    # when
    result = configurer.validate()

    # then
    assert result.has_errors()
    assert result.errors.__len__() == 5
    assert result.has_error('starts_at')
    assert result.has_error('expires_at')
    assert result.has_error('plan')
    assert result.has_error('user_limit')
    assert result.has_error('data')


def test_license_not_updated_because_same_exists(license_valid):
    # given
    gitlab = Mock()
    gitlab.get_license.return_value = {
        'starts_at': license_valid['starts_at'],
        'expires_at': license_valid['expires_at'],
        'plan': license_valid['plan'],
        'user_limit': license_valid['user_limit'],
    }
    configurer = LicenseConfigurer(gitlab, license_valid)

    # when
    configurer.configure()

    # then
    gitlab.get_license.assert_called_once()
    gitlab.set_license.assert_not_called()


def test_license_updated(license_valid):
    # given
    gitlab = Mock()
    gitlab.get_license.return_value = {
        'starts_at': '1900-01-01',
        'expires_at': license_valid['expires_at'],
        'plan': license_valid['plan'],
        'user_limit': license_valid['user_limit'],
    }
    configurer = LicenseConfigurer(gitlab, license_valid)

    # when
    configurer.configure()

    # then
    gitlab.get_license.assert_called_once()
    gitlab.set_license.assert_called_once_with(license_valid['data'])





import pytest

from gcasc import GitlabConfiguration


def test_error_raised_when_configuration_none():
    with pytest.raises(RuntimeError):
        GitlabConfiguration(None)


def test_error_raised_when_configuration_not_a_dict():
    with pytest.raises(RuntimeError):
        GitlabConfiguration('str')
